import os
import time
import math
import unittest
import datetime

from types import SimpleNamespace


class MyException(Exception):
    pass


class MyTestCase(unittest.TestCase):

    def setup(self):
        pass

    def setUp(self):
        self.setup()

    def tear_down(self):
        pass

    def tearDown(self):
        self.tear_down()


class MyTestArguments(SimpleNamespace):
    pass


class Logger(object):

    def __init__(self, prefix='log', to_dir='.', timestamp=True):
        self.f = None
        self.timestamp = timestamp
        os.makedirs(to_dir, exist_ok=True)
        now = datetime.datetime.now()
        if not prefix.endswith('_'):
            prefix = prefix + '_'
        file_name = '{}{}.txt'.format(prefix, now.strftime('%Y%m%d_%H%M%S'))
        self.f = open(os.path.join(to_dir, file_name), 'w')

    def print(self, message):
        now = datetime.datetime.now()
        if self.timestamp:
            message = '[' + now.strftime('%Y-%m-%d %H:%M:%S.%f') + '] ' + str(message)
        else:
            message = str(message)
        print(message)
        if self.f:
            self.f.write(message + '\n')

    def __del__(self):
        self.f.close()


def current_time_millis():
    return int(round(time.time() * 1000))


def current_time_secs():
    return int(round(current_time_millis() / 1000.0))


def elapsed_in_millis(start_time_millis):
    return current_time_millis() - start_time_millis


def elapsed_in_secs(start_time_secs):
    return current_time_secs() - start_time_secs


def duration(seconds):
    h = int(math.floor(seconds/3600.0))
    remainder = seconds - h * 3600
    m = int(math.floor(remainder/60.0))
    remainder = remainder - m * 60
    s = int(math.floor(remainder))
    return '{} hours, {} minutes, {} seconds'.format(h, m, s)



import os
import json

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session


class CastorClient:

    def __init__(self, client_id=None, client_secret=None):
        if client_id is None:
            with open('{}/castorclientid.txt'.format(os.environ['HOME']), 'r') as f:
                client_id = f.readline().strip()
        if client_secret is None:
            with open('{}/castorclientsecret.txt'.format(os.environ['HOME']), 'r') as f:
                client_secret = f.readline().strip()
        self.base_url = 'https://data.castoredc.com'
        self.token_url = self.base_url + '/oauth/token'
        self.api_url = self.base_url + '/api'
        self.session = self.create_session(client_id, client_secret, self.token_url)

    @staticmethod
    def create_session(client_id, client_secret, token_url):
        client = BackendApplicationClient(client_id=client_id)
        client_session = OAuth2Session(client=client)
        client_session.fetch_token(
            token_url=token_url,
            client_id=client_id,
            client_secret=client_secret,
        )
        return client_session

    def get_studies(self):
        response = self.session.get(self.api_url + '/study').json()
        studies = []
        for study in response['_embedded']['study']:
            studies.append(study)
        return studies

    def get_study_id(self, name):
        studies = self.get_studies()
        for study in studies:
            if study['name'] == name:
                return study['study_id']
        return None

    def get_fields(self, study_id, use_cache=True, verbose=False):
        if use_cache and os.path.isfile('cache/fields_{}.json'.format(study_id)):
            fields = json.load(open('cache/fields_{}.json'.format(study_id), 'r'))
            return fields
        else:
            url = self.api_url + '/study/{}/field'.format(study_id)
            response = self.session.get(url).json()
            page_count = response['page_count']
            fields = []
            for i in range(1, page_count + 1):
                url = self.api_url + '/study/{}/field?page={}'.format(study_id, i)
                response = self.session.get(url).json()
                for field in response['_embedded']['fields']:
                    fields.append(field)
                    if verbose:
                        print(field['id'])
            os.makedirs('cache', exist_ok=True)
            json.dump(fields, open('cache/fields_{}.json'.format(study_id), 'w'))
            return fields

    def get_records(self, study_id, use_cache=True, verbose=False):
        if use_cache and os.path.isfile('cache/records_{}.json'.format(study_id)):
            records = json.load(open('cache/records_{}.json'.format(study_id), 'r'))
            return records
        else:
            url = self.api_url + '/study/{}/record'.format(study_id)
            response = self.session.get(url).json()
            page_count = response['page_count']
            records = []
            for i in range(1, page_count + 1):
                url = self.api_url + '/study/{}/record?page={}'.format(study_id, i)
                response = self.session.get(url).json()
                for record in response['_embedded']['records']:
                    if not record['id'].startswith('ARCHIVED'):
                        records.append(record)
                        if verbose:
                            print(record['id'])
            os.makedirs('cache', exist_ok=True)
            json.dump(records, open('cache/records_{}.json'.format(study_id), 'w'))
            return records

    def get_field_data(self, study_id, record_id, field_id):
        url = self.api_url + '/study/{}/record/{}/study-data-point/{}'.format(study_id, record_id, field_id)
        field_data = self.session.get(url).json()
        return field_data


if __name__ == '__main__':
    client = CastorClient()
    print(client.get_studies())



import os
import h5py
import random
import pydicom
import argparse
import numpy as np

from types import SimpleNamespace

from barbell2.utils import Logger
from barbell2.lib.dicom import is_dicom
from barbell2.lib.dicom import Dcm2Numpy, Tag2NumPy

from sklearn.model_selection import train_test_split


class CreateH5:

    def __init__(self, args):
        if isinstance(args, dict):
            self.args = SimpleNamespace(**args)
        else:
            self.args = args
        self.logger = Logger(
            prefix='createh5', to_dir=self.args.log_dir)

    @staticmethod
    def has_dimensions(dcm_file, rows, columns):
        p = pydicom.read_file(dcm_file)
        if int(p.Rows) == rows and int(p.Columns) == columns:
            return True
        return False

    @staticmethod
    def get_tag_file_path(dcm_file, append=False):
        if dcm_file.endswith('.dcm') and not append:
            return dcm_file[:-4] + '.tag'
        else:
            return dcm_file + '.tag'

    def collect_files(self, data_dir, rows, columns, is_train):
        count = 1
        files_list = []
        for root, dirs, files in os.walk(data_dir):
            for f in files:
                f = os.path.join(root, f)
                if is_dicom(f):
                    if self.has_dimensions(f, rows, columns):
                        if is_train:
                            # Try to get the TAG file path. If we can't find it, try not
                            # replacing the 'dcm' extension with '.tag' but simply append to it
                            ok = False
                            tag_file = self.get_tag_file_path(f)
                            if os.path.isfile(tag_file):
                                ok = True
                            else:
                                tag_file = self.get_tag_file_path(f, append=True)
                                if os.path.isfile(tag_file):
                                    ok = True
                            if ok:
                                files_list.append((f, tag_file))
                                self.logger.print('{:03d} Added {} and TAG file'.format(count, f))
                                count += 1
                            else:
                                self.logger.print('File {} does not have corresponding TAG file'.format(f))
                        else:
                            files_list.append((f, ))
                            self.logger.print('{:03d} Added {}'.format(count, f))
                    else:
                        self.logger.print('File {} has wrong dimensions'.format(f))
                else:
                    pass
        return files_list

    @staticmethod
    def shuffle_files(files):
        random.shuffle(files)
        return files

    @staticmethod
    def split_files(files, test_size):
        if test_size > 0.0:
            return train_test_split(files, test_size=test_size)
        return files, []

    @staticmethod
    def get_dcm_pixels(file_path):
        dcm2numpy = Dcm2Numpy()
        dcm2numpy.set_input_dicom_file_path(file_path)
        dcm2numpy.set_normalize_enabled()  # To ensure true HU values
        dcm2numpy.execute()
        return dcm2numpy.get_output_numpy_array()

    @staticmethod
    def get_tag_pixels(file_path, shape):
        tag2numpy = Tag2NumPy(shape)
        tag2numpy.set_input_tag_file_path(file_path)
        tag2numpy.execute()
        pixels = tag2numpy.get_output_numpy_array()
        # Note that sometimes pixels = None because of failure to reshape the pixel array
        # to the size of the DICOM image
        return pixels

    @staticmethod
    def labels_ok(labels):
        for label in [0, 1, 2, 3]:
            if label not in labels:
                return False
        return len(labels) == 4

    def update_labels(self, pixels):
        pixels[pixels == 2] = 0
        pixels[pixels == 4] = 0
        pixels[pixels == 11] = 0
        pixels[pixels == 12] = 0
        pixels[pixels == 14] = 0
        pixels[pixels == 5] = 2
        pixels[pixels == 7] = 3
        labels = np.unique(pixels)
        if self.labels_ok(labels):
            return pixels, labels
        return None, None

    @staticmethod
    def get_file_id(file_path):
        items = os.path.split(file_path)
        file_id = os.path.splitext(items[1])[0]
        return file_id

    def create_file(self, files, output_file, is_train):
        with h5py.File(output_file, 'w') as h5f:
            count = 1
            for file_pair in files:
                file_id = self.get_file_id(file_pair[0])
                dcm_pixels = self.get_dcm_pixels(file_pair[0])
                if dcm_pixels is None:
                    self.logger.print('DICOM file {} has problems with pixel data'.format(file_pair[0]))
                    continue
                if is_train:
                    tag_pixels = self.get_tag_pixels(file_pair[1], dcm_pixels.shape)
                    if tag_pixels is None:
                        self.logger.print('TAG file {} cannot be reshaped to right dimensions'.format(file_pair[1]))
                        continue
                    tag_pixels, unique_labels = self.update_labels(tag_pixels)
                    if tag_pixels is None:
                        self.logger.print('TAG file {} has wrong labels {}'.format(file_pair[1], unique_labels))
                        continue
                    group = h5f.create_group('{}'.format(file_id))
                    group.create_dataset('image', data=dcm_pixels)
                    group.create_dataset('labels', data=tag_pixels)
                    self.logger.print('{:04d} added images and labels for file ID {}'.format(count, file_id))
                    count += 1
                else:
                    group = h5f.create_group('{}'.format(file_id))
                    group.create_dataset('images', data=dcm_pixels)

    def execute(self):
        is_train = True if self.args.type == 'train' else False
        if is_train:
            self.logger.print('Collecting files for training')
        else:
            self.logger.print('Collecting files for prediction')
        files_list = self.collect_files(
            self.args.data_dir, self.args.rows, self.args.columns, is_train)
        if is_train and self.args.test_size > 0.0:
            self.logger.print('Shuffling files before training/test split')
            files_list = self.shuffle_files(files_list)
        if is_train and self.args.test_size > 0.0:
            self.logger.print('Splitting files with test size of {}%'.format(int(100 * self.args.test_size)))
        training_files, test_files = self.split_files(files_list, self.args.test_size)
        self.logger.print('Training files: {}'.format(len(training_files)))
        if is_train and self.args.test_size > 0.0:
            self.logger.print('Test files: {}'.format(len(test_files)))
        if is_train and self.args.test_size > 0.0:
            self.logger.print('Creating H5 file for testing')
            self.create_file(test_files, 'test.h5', is_train)
        if is_train:
            self.logger.print('Creating H5 file for training')
            self.create_file(training_files, 'train.h5', is_train)
        else:
            self.logger.print('Creating H5 file for prediction')
            self.create_file(training_files, 'predict.h5', is_train)
        self.logger.print('Done')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'data_dir',
        help='Directory containing DICOM and (optionally) TAG files',
    )
    parser.add_argument(
        '--rows',
        help='Number of pixel rows in each image, same as image height (default: 512)',
        type=int, default=512,
    )
    parser.add_argument(
        '--columns',
        help='Number of pixel columns in each image, same as image width (default: 512)',
        type=int, default=512,
    )
    parser.add_argument(
        '--type',
        help='Which type of H5 file to create. One of [train, predict] (default: train)',
        default='train',
    )
    parser.add_argument(
        '--test_size',
        help='Percentage of data to use for testing (between 0 and 1, default: 0.2)',
        type=float, default=0.2,
    )
    parser.add_argument(
        '--log_dir',
        help='Directory where to write logging info (default: .)',
        default='.',
    )
    args = parser.parse_args()
    x = CreateH5(args)
    x.execute()


if __name__ == '__main__':
    main()



import os
import json
import numpy as np
import SimpleITK as sitk


class Dcm2Masks(object):

    def __init__(self):
        self._input_dicom_file_path = None
        self._label_map = None
        self._output_dir = None
        self._output_file_paths = None
        self._overwrite_output = False

    # INTERFACE

    def set_input_dicom_file_path(self, file_path):
        self._input_dicom_file_path = file_path

    def set_label_map(self, label_map):
        self._label_map = label_map

    def set_label_map_tomovision(self):
        self._label_map = {1: 'muscle', 5: 'SAT', 7: 'VAT'}

    def set_output_dir(self, output_dir):
        self._output_dir = output_dir

    def set_overwrite_output(self, value):
        self._overwrite_output = value

    def get_output_file_paths(self):
        return self._output_file_paths

    # INTERNAL METHODS

    @staticmethod
    def _get_int16_label_map(label_map):
        lm = {}
        for label in label_map.keys():
            label_int16 = np.int16(label)
            lm[label_int16] = label_map[label]
        return lm

    @staticmethod
    def _threshold_image(image, label):
        image = sitk.BinaryThreshold(image, float(label), float(label))
        return image

    @staticmethod
    def extract_pixels_by_label(pixels, label):
        pixels_copy = np.copy(pixels)
        pixels_copy[pixels_copy != label] = np.uint16(0)
        pixels_copy[pixels_copy == label] = np.uint16(1)
        return pixels_copy

    def execute(self):
        os.makedirs(self._output_dir, exist_ok=True)
        file_base_name = os.path.splitext(self._input_dicom_file_path)[0]
        label_map = self._get_int16_label_map(self._label_map)
        file_paths = {}
        for label in label_map.keys():

            # f = open(self._input_dicom_file_path, 'rb')
            # p = pydicom.read_file(f)
            # f.close()
            # pixels = p.pixel_array.flat
            # pixels = self.extract_pixels_by_label(pixels, label)
            # p.PixelData = pixels.tobytes()
            # file_path = os.path.join(self._output_dir, '{}_{}.dcm'.format(file_base_name, label_map[label]))
            # try:
            #     os.stat(file_path)
            #     if not self._overwrite_output:
            #         print('Error: file {} already exists!'.format(file_path))
            #         return
            # except FileNotFoundError:
            #     pass
            # p.save_as(file_path)
            # file_paths[label_map[label]] = file_path

            reader = sitk.ImageFileReader()
            reader.SetImageIO('GDCMImageIO')
            reader.SetFileName(self._input_dicom_file_path)
            image = reader.Execute()
            image_new = self._threshold_image(image, label)
            writer = sitk.ImageFileWriter()
            file_path = os.path.join(self._output_dir, '{}_{}.dcm'.format(file_base_name, label_map[label]))
            try:
                os.stat(file_path)
                if not self._overwrite_output:
                    print('Error: file {} already exists!'.format(file_path))
                    return
            except FileNotFoundError:
                pass
            writer.SetFileName(file_path)
            file_paths[label_map[label]] = file_path
            writer.SetImageIO('GDCMImageIO')
            writer.Execute(image_new)

        self._output_file_paths = file_paths


if __name__ == '__main__':
    node = Dcm2Masks()
    node.set_input_dicom_file_path('/Volumes/USB_SECURE1/data/radiomics/projects/004_ovarium/data/test/out/IM_1_tag.dcm')
    node.set_label_map_tomovision()
    node.set_output_dir('/Volumes/USB_SECURE1/data/radiomics/projects/004_ovarium/data/test/out')
    node.set_overwrite_output(True)
    node.execute()
    print(json.dumps(node.get_output_file_paths(), indent=4))



import os

from .dcm2masks import Dcm2Masks
from .dcm2nifti import Dcm2Nifti
from .dcm2numpy import Dcm2Numpy
from .nifti2masks import Nifti2Masks
from .tag2dcm import Tag2Dcm
from .tag2nifti import Tag2Nifti
from .tag2numpy import Tag2NumPy


def is_dicom(file_path):
    if not os.path.isfile(file_path):
        return False
    if file_path.startswith('._'):
        return False
    try:
        with open(file_path, "rb") as f:
            return f.read(132).decode("ASCII")[-4:] == "DICM"
    except UnicodeDecodeError:
        return False


def is_tag(file_path):
    return file_path.endswith('.tag') and not file_path.startswith('._')



import os
import h5py
import random
import pydicom
import numpy as np

from barbell2.utils import Logger
from barbell2.lib.dicom import is_dicom
from barbell2.lib.dicom import Dcm2Numpy, Tag2NumPy

from sklearn.model_selection import train_test_split


class CreateHDF5:

    def __init__(self, dir_path, output_files, rows, columns, test_size=0.0, is_training=True, log_dir='.'):
        self.dir_path = dir_path
        self.is_training = is_training
        if isinstance(output_files, str):
            self.output_files = [output_files]
        else:
            self.output_files = output_files
        if not self.is_training:
            if len(self.output_files) != 1:
                raise RuntimeError('For prediction provide only one output file')
        self.rows = rows
        self.columns = columns
        self.test_size = test_size
        os.makedirs(log_dir, exist_ok=True)
        self.log = Logger(prefix='createh5_', to_dir=log_dir)

    @staticmethod
    def has_dimensions(dcm_file, rows, columns):
        p = pydicom.read_file(dcm_file)
        if int(p.Rows) == rows and int(p.Columns) == columns:
            return True
        return False

    @staticmethod
    def get_tag_file_path(dcm_file, append=False):
        if dcm_file.endswith('.dcm') and not append:
            return dcm_file[:-4] + '.tag'
        else:
            return dcm_file + '.tag'

    def collect_files(self, dir_path, rows, columns, is_training):
        files_list = []
        count = 1
        for root, dirs, files in os.walk(dir_path):
            for f in files:
                f = os.path.join(root, f)
                if is_dicom(f):
                    if self.has_dimensions(f, rows, columns):
                        if is_training:
                            append = False
                            tag_file = self.get_tag_file_path(f)
                            if os.path.isfile(tag_file):
                                append = True
                            else:
                                tag_file = self.get_tag_file_path(f, append=True)
                                if os.path.isfile(tag_file):
                                    append = True
                                else:
                                    self.log.print('File {} does not have TAG file'.format(f))
                            if append:
                                files_list.append((f, tag_file))
                                self.log.print('{:03d} Appended {} and TAG file'.format(count, f))
                                count += 1
                        else:
                            files_list.append((f, ))
                            self.log.print('{:03d} Appended {}'.format(count, f))
                            count += 1
                    else:
                        self.log.print('File {} has wrong dimensions'.format(f))
                else:
                    pass
        return files_list

    @staticmethod
    def shuffle_files(files):
        random.shuffle(files)
        return files

    @staticmethod
    def split_files(files, test_size):
        if test_size > 0.0:
            return train_test_split(files, test_size=test_size)
        return files, []

    @staticmethod
    def get_dcm_pixels(file_path):
        dcm2numpy = Dcm2Numpy()
        dcm2numpy.set_input_dicom_file_path(file_path)
        dcm2numpy.set_normalize_enabled()  # To ensure true HU values
        dcm2numpy.execute()
        return dcm2numpy.get_output_numpy_array()

    @staticmethod
    def get_tag_pixels(file_path, shape):
        tag2numpy = Tag2NumPy(shape)
        tag2numpy.set_input_tag_file_path(file_path)
        tag2numpy.execute()
        pixels = tag2numpy.get_output_numpy_array()
        # Note that sometimes pixels = None because of failure to reshape the pixel array
        # to the size of the DICOM image
        return pixels

    @staticmethod
    def labels_ok(labels):
        for label in [0, 1, 2, 3]:
            if label not in labels:
                return False
        return len(labels) == 4

    def update_labels(self, pixels):
        pixels[pixels == 2] = 0
        pixels[pixels == 4] = 0
        pixels[pixels == 11] = 0
        pixels[pixels == 12] = 0
        pixels[pixels == 14] = 0
        pixels[pixels == 5] = 2
        pixels[pixels == 7] = 3
        labels = np.unique(pixels)
        if self.labels_ok(labels):
            return pixels, labels
        return None, None

    @staticmethod
    def get_file_id(file_path):
        items = os.path.split(file_path)
        file_id = os.path.splitext(items[1])[0]
        return file_id

    def create_file(self, output_file, files):
        with h5py.File(output_file, 'w') as h5f:
            count = 1
            for file_pair in files:
                file_id = self.get_file_id(file_pair[0])
                dcm_pixels = self.get_dcm_pixels(file_pair[0])
                if dcm_pixels is None:
                    self.log.print('DICOM file {} has problems with pixel data'.format(file_pair[0]))
                    continue
                if self.is_training:
                    tag_pixels = self.get_tag_pixels(file_pair[1], dcm_pixels.shape)
                    if tag_pixels is None:
                        self.log.print('TAG file {} cannot be reshaped to right dimensions'.format(file_pair[1]))
                        continue
                    tag_pixels, unique_labels = self.update_labels(tag_pixels)
                    if tag_pixels is None:
                        self.log.print('TAG file {} has wrong labels {}'.format(file_pair[1], unique_labels))
                        continue
                    group = h5f.create_group('{}'.format(file_id))
                    group.create_dataset('image', data=dcm_pixels)
                    group.create_dataset('labels', data=tag_pixels)
                    self.log.print('{:04d} added images and labels for file ID {}'.format(count, file_id))
                    count += 1
                else:
                    group = h5f.create_group('{}'.format(file_id))
                    group.create_dataset('images', data=dcm_pixels)
            self.log.print('Done')

    def create_hdf5(self):
        if self.is_training:
            self.log.print('Collecting files for training')
        else:
            self.log.print('Collecting files for prediction')
        files = self.collect_files(self.dir_path, self.rows, self.columns, self.is_training)
        if self.is_training and self.test_size > 0.0:
            self.log.print('Shuffling files before training/test split')
            files = self.shuffle_files(files)
        if self.test_size < 1.0:
            self.log.print('Splitting files with test size of {}%'.format(int(100 * self.test_size)))
        training_files, test_files = self.split_files(files, self.test_size)
        self.log.print('Training files: {}'.format(len(training_files)))
        self.log.print('Test files: {}'.format(len(test_files)))
        self.log.print('Creating H5 training file')
        self.create_file(self.output_files[0], training_files)
        if len(self.output_files) == 2:
            self.log.print('Creating H5 test file')
            self.create_file(self.output_files[1], test_files)
            return self.output_files[0], self.output_files[1]
        else:
            return self.output_files[0], None



import os
import pytest

from barbell2.createh5 import CreateH5

base_dir = '{}/data/surfdrive/projects/20210203_autosegl3'.format(os.environ['HOME'])


def clean_up():
    if os.path.isfile('train.h5'):
        os.remove('train.h5')
    if os.path.isfile('test.h5'):
        os.remove('test.h5')
    if os.path.isfile('predict.h5'):
        os.remove('predict.h5')


def test_train_test():
    """ Both DICOM and TAG files should be collected and the data split up in a training and test set. """
    clean_up()
    d = os.path.join(base_dir, 'dicom_and_tag')
    args = {
        'data_dir': d,
        'rows': 512, 'columns': 512,
        'type': 'train',
        'test_size': 0.2,
        'log_dir': '.',
    }
    x = CreateH5(args)
    x.execute()
    assert os.path.isfile('train.h5')
    assert os.path.isfile('test.h5')
    clean_up()


def test_train():
    """ Both DICOM and TAG files should be collected but all data used for training. """
    clean_up()
    d = os.path.join(base_dir, 'dicom_and_tag')
    args = {
        'data_dir': d,
        'rows': 512, 'columns': 512,
        'type': 'train',
        'test_size': 0.0,
        'log_dir': '.',
    }
    x = CreateH5(args)
    x.execute()
    assert os.path.isfile('train.h5')
    clean_up()


def test_predict():
    """ Only DICOM files should be collected for prediction of labels. """
    clean_up()
    d = os.path.join(base_dir, 'dicom')
    args = {
        'data_dir': d,
        'rows': 512, 'columns': 512,
        'type': 'predict',
        'test_size': 0.0,
        'log_dir': '.',
    }
    x = CreateH5(args)
    x.execute()
    assert os.path.isfile('predict.h5')
    clean_up()



import json
import pandas as pd
import numpy as np


class CastorExportClient:

    def __init__(self, show_params=True):
        """
        Constructs instance of this class.
        :param show_params: Whether to print parameters or not (default True)
        """
        self.params = {
            'sheet_name_data': 'Study results',                 # with spaces
            'sheet_name_data_dict': 'Study variable list',      # ''
            'sheet_name_data_options': 'Field options',         # ''
            'data_dict_crf_name': 'Step_name',                  # without spaces because column names are updated
            'data_dict_var_name': 'Variable_name',              # ''
            'data_dict_field_type': 'Field_type',               # ''
            'data_dict_field_label': 'Field_label',             # ''
            'data_dict_option_group_name': 'Optiongroup_name',  # ''
            'data_options_group_name': 'Option_group_name',     # ''
            'data_options_name': 'Option_name',                 # ''
            'data_options_value': 'Option_value',               # ''
            'data_cols_ignore': [
                'Record_Id',                                    # without spaces because column names are updated
                'Institute_Abbreviation',                       # ''
                'Record_Creation_Date',                         # ''
            ],
            'patient_id_field_name': 'dpca_idcode',
            'surgery_date_field_name': 'dpca_datok',
            'data_miss_float': [999, 9999, 99999, 999.0, 9999.0],
            'data_miss_date': ['09-09-1809'],
            'to_pandas': {
                'dropdown': 'Int64',
                'radio': 'Int64',
                'string': 'object',
                'textarea': 'object',
                'remark': 'object',
                'date': 'datetime64[ns]',
                'year': 'Int64',
                'numeric': 'float64',
            }
        }

        if show_params:
            print(json.dumps(self.params, indent=4))

        self.data = None
        self.data_dict = {}
        self.data_options = {}

    @staticmethod
    def remove_spaces(value):
        """
        Removes spaces from given value and replaces them with underscores.
        :param value: Value to process
        :return: Processed value
        """
        return value.replace(' ', '_')

    def to_pandas_type(self, field_type):
        """
        Lookup Pandas data type for given Castor field type.
        :param field_type: Castor field type name
        :return: Corresponding Pandas data type or None if not found
        """
        if field_type in list(self.params['to_pandas'].keys()):
            return self.params['to_pandas'][field_type]
        return None

    def load_data(self, file_path, verbose=False):
        """
        Loads Castor export Excel file containing the data, data dictionary and field options.
        :param file_path: Path to Excel file
        :param verbose: Verbose
        :return: Tuple of data, data dictionary and options
        """
        # Load data dictionary first and remove spaces from columns
        df_data_dict = pd.read_excel(file_path, sheet_name=self.params['sheet_name_data_dict'], dtype='object')
        df_data_dict.columns = map(self.remove_spaces, df_data_dict.columns)

        # Fill missing values with np.nan
        for column in df_data_dict.columns:
            df_data_dict[column] = df_data_dict[column].fillna(np.nan)

        data_dict = {}

        # Add columns to ignore to data dictionary
        for column in self.params['data_cols_ignore']:
            data_dict[column] = {
                'crf_name': '',
                'field_label': column,
                'field_type': 'string',
                'pandas_type': 'object',
                'option_group_name': None,
            }

        # Store field definitions in data dictionary
        for idx, row in df_data_dict.iterrows():
            var_name = row[self.params['data_dict_var_name']]
            if var_name is None or var_name == '' or pd.isna(var_name):
                continue
            data_dict[var_name] = {
                'crf_name': row[self.params['data_dict_crf_name']],
                'field_label': row[self.params['data_dict_field_label']],
                'field_type': row[self.params['data_dict_field_type']],
                'pandas_type': self.to_pandas_type(row[self.params['data_dict_field_type']]),
                'option_group_name': row[self.params['data_dict_option_group_name']],
            }

        self.data_dict = data_dict

        # Load data and remove spaces from columns
        df_data = pd.read_excel(file_path, sheet_name=self.params['sheet_name_data'])
        df_data.columns = map(self.remove_spaces, df_data.columns)

        for column in df_data.columns:

            if column not in self.data_dict.keys():
                continue

            # Check Pandas type
            pandas_type = self.to_pandas_type(self.data_dict[column]['field_type'])

            # Fill with NaN values and create new series according to Pandas type
            df_data[column] = df_data[column].fillna(np.nan)
            df_data[column] = pd.Series(data=df_data[column], dtype=pandas_type)

            # Missing values for floats en dates are specific to Castor data. Let's
            # replace these with either np.nan or pd.NaT
            if pandas_type == 'float64':
                for mv in self.params['data_miss_float']:
                    df_data.loc[df_data[column] == mv, column] = np.nan
            elif pandas_type == 'datetime64[ns]':
                for mv in self.params['data_miss_date']:
                    df_data.loc[df_data[column] == mv, column] = pd.NaT
            else:
                pass

            if verbose:
                print('Processed column {}'.format(column))

        self.data = df_data

        # Load options and remove spaces from columns
        df_data_options = pd.read_excel(file_path, sheet_name=self.params['sheet_name_data_options'], dtype='object')
        df_data_options.columns = map(self.remove_spaces, df_data_options.columns)

        # Fill in missing values
        for column in df_data_options.columns:
            df_data_options[column] = df_data_options[column].fillna(np.nan)

        data_options = {}

        for idx, row in df_data_options.iterrows():

            # Store options in option dictionary
            option_group = row[self.params['data_options_group_name']]
            if option_group not in list(data_options.keys()):
                data_options[option_group] = []
            data_options[option_group].append(
                (int(row[self.params['data_options_value']]), row[self.params['data_options_name']]))

        self.data_options = data_options

        return self.data, self.data_dict, self.data_options

    def find_option_group(self, text=''):
        """
        Finds option groups and corresponding option values for the given (partial) text.
        :param text: (Part of) option name or group name (default='' returns all options groups)
        """
        option_groups = {}
        for option_group, options in self.data_options.items():
            if text.lower() in option_group.lower():
                option_groups[option_group] = options
                continue
            for option in options:
                if text.lower() in option[1].lower():
                    option_groups[option_group] = options
        return option_groups

    def find_variable(self, keys):
        """
        Finds variable definitions that contain <text> in either the name or label. Info returned
        contains: CRF name, field label, field type, Pandas type and option group name (if applicable).
        :param keys: Key or list of keys
        :return: List of variable definitions matching given keys
        """
        if isinstance(keys, str):
            keys = [keys]
        if not isinstance(keys, list):
            print('Keys must be string or list of strings')
        definitions = []
        for name, definition in self.data_dict.items():
            found = False
            for key in keys:
                if key.lower() in name.lower():
                    found = True
                elif key.lower() in definition['field_label'].lower():
                    found = True
                elif key.lower() in definition['crf_name'].lower():
                    found = True
                else:
                    pass
                if found:
                    if not pd.isna(definition['option_group_name']):
                        option_group = self.find_option_group(definition['option_group_name'])
                        definition['options'] = option_group[definition['option_group_name']]
                    definitions.append((name, definition))
        return definitions

    def find_values(self, var_name):
        """
        Finds values for given <var_name>.
        :param var_name: Variable (column) to display values for.
        :return: Pandas series with values
        """
        return self.data[var_name]

    def find_missing(self, in_column, show_columns):
        """
        Finds records with missing values in column <in_column>. Each record is displayed as indicated
        by <show_columns>
        :param in_column: Column that is searched for missing values.
        :param show_columns: List of column names to show when displaying records with missing values.
        :return:
        """
        if isinstance(show_columns, str):
            if not show_columns in self.data.columns:
                print('Column {} not found'.format(show_columns))
                return None
        elif isinstance(show_columns, list) or isinstance(show_columns, tuple):
            for column in show_columns:
                if not column in self.data.columns:
                    print('Column {} not found'.format(show_columns))
                    return None
        else:
            print('Wrong type for show_columns, must be string, list or tuple')
            return None
        missing = self.data[in_column].isnull()
        return self.data.loc[missing == True, show_columns]

    def find_duplicate_records(self, columns):
        """
        Finds duplicate records in the export file based on the given key columns.
        :return: Dictionary with keys that contain more than 1 record.
        """
        for column in columns:
            if not column in self.data.columns:
                print('Could not find column {}'.format(column))
                return {}
        record_counts = {}
        for idx, row in self.data.iterrows():
            key = []
            for column in columns:
                key_item = row[column]
                if self.data[column].dtype == 'datetime64[ns]':
                    key_item = '{}-{}-{}'.format(key_item.year, key_item.month, key_item.day)
                key.append(key_item)
            key = tuple(key)
            if key not in record_counts.keys():
                record_counts[key] = 0
            record_counts[key] += 1
        duplicates = {}
        for key, count in record_counts.items():
            if count > 1:
                duplicates[key] = count
        return duplicates

    def query(self, query_string):
        """
        Run the given query on the Pandas dataframe. If query fails with a value error we try again,
        using the 'python' engine. This seems only to happen when running the query from within
        a Jupyter notebook.

        Example queries:
        ----------------
        Given the column (variable) names 'dob' and 'gender', select all records where dob is earlier than
        01-01-1945 and gender is male. Note: you must lookup the option value for 'male' using the find_option_group
        function, e.g., find_option_group('gender')

            client.query('dob < "01-01-1945" & gender == 1')

        Common query variables:
        -----------------------

        DPCA:
        - dpca_gebdat       Birth date
        - dpca_gebjaar      Year of birth
        - dpca_geslacht     Gender (1: male, 2: female)
        - dpca_typok        Type of surgery procedure
        - dpca_comorb       Comorbidities (0: no, 1: yes)
        - dpca_ovl          Deceased (0: no, 1: yes)
        - dpca_datovl       Date of death
        - dpca_afwijking    Most recent disease diagnosis

        DHBA:
        - dhba_diagnose     Most recent disease diagnosis

        :param query_string: Query string
        :return: Result data frame
        """
        try:
            return self.data.query(query_string)
        except ValueError:
            return self.data.query(query_string, engine='python')


if __name__ == '__main__':
    CastorExportClient()
